@@ -4,7 +4,7 @@
 # Univention Management Console
 #  Univention Directory Manager Module
 #
-# Copyright 2019-2021 Univention GmbH
+# Copyright 2019-2022 Univention GmbH
 #
 # https://www.univention.de/
 #
@@ -37,6 +37,7 @@ from __future__ import print_function
 from __future__ import unicode_literals
 
 import os
+import atexit
 import time
 import locale
 import signal
@@ -45,28 +46,37 @@ import argparse
 import traceback
 import logging
 
+from setproctitle import getproctitle, setproctitle
+
 import tornado.log
 import tornado.ioloop
 from tornado.httpserver import HTTPServer
-from tornado.netutil import bind_unix_socket
+from tornado.netutil import bind_sockets, bind_unix_socket
 
 # IMPORTANT NOTICE: we must import as few modules as possible, so that univention.admin is not yet imported
 # because importing the UDM handlers would cause that the gettext translation gets applied before we set a locale
 from univention.management.console.config import ucr
 from univention.management.console.log import log_init, log_reopen, CORE
 from univention.lib.i18n import Locale
+from univention.admin.rest.shared_memory import shared_memory
+
+try:
+	from multiprocessing.util import _exit_function
+except ImportError:
+	_exit_function = None
+
+proctitle = getproctitle()
 
 
 class Server(object):
 
-	def start(self, args):
-		if os.fork() > 0:
-			os._exit(0)
-		self.run()
+	child_id = None
 
 	def run(self, args):
+		self.child_id = None
+		setproctitle(proctitle + '   # server')
 		# locale must be set before importing UDM!
-		log_init('/dev/stdout', args.debug)
+		log_init('/dev/stdout', args.debug, args.processes != 1)
 		language = str(Locale(args.language))
 		locale.setlocale(locale.LC_MESSAGES, language)
 		os.umask(0o077)  # FIXME: should probably be changed, this is what UMC sets
@@ -76,38 +86,89 @@ class Server(object):
 		os.environ['PATH'] = '/bin:/sbin:/usr/bin:/usr/sbin'
 		os.environ['LANG'] = language
 
-		import univention.admin.modules as udm_modules
-		udm_modules.update()
-
-		from univention.admin.rest.module import Application
-		application = Application(serve_traceback=ucr.is_true('directory/manager/rest/show-tracebacks', True))
-
-		server = HTTPServer(application)
-		if args.port:
-			server.bind(args.port)
-		server.start(args.cpus)
-
-		if args.unix_socket:
-			socket = bind_unix_socket(args.unix_socket)
-			server.add_socket(socket)
-		signal.signal(signal.SIGTERM, partial(self.signal_handler_stop, server))
-		signal.signal(signal.SIGINT, partial(self.signal_handler_stop, server))
-		signal.signal(signal.SIGHUP, self.signal_handler_reload)
-
+		# tornado logging
 		channel = logging.StreamHandler()
 		channel.setFormatter(tornado.log.LogFormatter(fmt='%(color)s%(asctime)s  %(levelname)10s      (%(process)9d) :%(end_color)s %(message)s', datefmt='%d.%m.%y %H:%M:%S'))
 		logger = logging.getLogger()
 		logger.setLevel(logging.INFO)
 		logger.addHandler(channel)
 
+		# start sharing memory (before fork, before first usage, after import)
+		shared_memory.start()
+
+		import univention.admin.modules as udm_modules
+		udm_modules.update()
+
+		# bind sockets
+		socks = []
+		if args.port:
+			socks.extend(bind_sockets(args.port, '127.0.0.1', reuse_port=True))
+
+		if args.unix_socket:
+			socks.append(bind_unix_socket(args.unix_socket))
+
+		# signal handers
+		signal.signal(signal.SIGTERM, partial(self.signal_handler_stop, None))
+		signal.signal(signal.SIGINT, partial(self.signal_handler_stop, None))
+		signal.signal(signal.SIGHUP, self.signal_handler_reload)
+
+		# start mutliprocessing
+		if args.processes != 1:
+			if _exit_function is not None:
+				atexit.unregister(_exit_function)
+			self.socks = socks
+			try:
+				child_id = tornado.process.fork_processes(args.processes, 0)
+			except RuntimeError as exc:  # tornados way to exit from multiprocessing on failures
+				CORE.info('Stopped process: %s' % (exc,))
+				self.signal_handler_stop(None, signal.SIGTERM, None)
+			else:
+				self.start_child(child_id)
+				CORE.info('process ended')
+		else:
+			self.run_server(socks)
+
+	def start_child(self, child_id):
+		setproctitle(proctitle + '   # child %s' % (child_id,))
+		self.child_id = child_id
+		CORE.info('Started child %s' % (self.child_id,))
+		shared_memory.children[self.child_id] = os.getpid()
+		self.run_server(self.socks)
+
+	def run_server(self, socks):
+		from univention.admin.rest.module import Application
+		application = Application(serve_traceback=ucr.is_true('directory/manager/rest/show-tracebacks', True))
+
+		server = HTTPServer(application)
+		server.add_sockets(socks)
+
+		# signal handers (after forking)
+		signal.signal(signal.SIGTERM, partial(self.signal_handler_stop, server))
+		signal.signal(signal.SIGINT, partial(self.signal_handler_stop, server))
+		signal.signal(signal.SIGHUP, self.signal_handler_reload)
+
 		try:
 			tornado.ioloop.IOLoop.current().start()
 		except Exception:
 			CORE.error(traceback.format_exc())
+			self.signal_handler_stop(server, signal.SIGTERM, None)
 			raise
 
 	def signal_handler_stop(self, server, sig, frame):
-		io_loop = tornado.ioloop.IOLoop.instance()
+		if self.child_id is None:
+			try:
+				children_pids = list(shared_memory.children.values())
+			except Exception:  # multiprocessing failure
+				children_pids = []
+			CORE.info('stopping children: %r' % (children_pids,))
+			for pid in children_pids:
+				self.safe_kill(pid, sig)
+
+			shared_memory.shutdown()
+		else:
+			CORE.info('shutting down in one second')
+
+		io_loop = tornado.ioloop.IOLoop.current()
 		loop = getattr(io_loop, 'asyncio_loop', io_loop)  # Support Python2+3 Tornado version
 
 		def stop_loop(deadline):
@@ -119,37 +180,40 @@ class Server(object):
 
 		def shutdown():
 			# wait one second to shutdown
-			server.stop()
+			if server:
+				server.stop()
 			stop_loop(time.time() + 1)
 
 		io_loop.add_callback_from_signal(shutdown)
 
 	def signal_handler_reload(self, signal, frame):
+		if self.child_id is None:
+			for pid in shared_memory.children.values():
+				self.safe_kill(pid, signal)
 		ucr.load()
 		log_reopen()
 
+	def safe_kill(self, pid, signo):
+		try:
+			os.kill(pid, signo)
+		except EnvironmentError as exc:
+			CORE.error('Could not kill(%s) %s: %s' % (signo, pid, exc))
+		else:
+			os.waitpid(pid, os.WNOHANG)
+
 	@classmethod
 	def main(cls):
 		server = cls()
 
 		parser = argparse.ArgumentParser()
-		parser.add_argument('-d', '--debug', type=int, default=int(ucr.get('directory/manager/rest/debug/level', 2)), help='debug level')
+		parser.add_argument('-d', '--debug', type=int, default=ucr.get_int('directory/manager/rest/debug/level'), help='debug level')
 		parser.add_argument('-l', '--language', default='C', help='The process locale')
 		parser.add_argument('-s', '--unix-socket', help='Bind to a UNIX socket')
 		parser.add_argument('-p', '--port', help='Bind to a TCP port')
-		parser.add_argument('-c', '--cpus', type=int, default=int(ucr.get('directory/manager/rest/cpus', 1)), help='How many processes should be forked')
-
-		subparsers = parser.add_subparsers(title='actions', description='All available actions')
-
-		start_parser = subparsers.add_parser('start', description='Start the service')
-		start_parser.set_defaults(func=server.start)
-
-		run_parser = subparsers.add_parser('run', description='Start the service in foreground')
-		run_parser.set_defaults(func=server.run)
+		parser.add_argument('-c', '--processes', type=int, default=ucr.get_int('directory/manager/rest/processes'), help='How many processes should be forked')
 
 		args = parser.parse_args()
-
-		args.func(args)
+		server.run(args)
 
 
 if __name__ == "__main__":
